#ifndef UTIL_LINUX_LSFD_DECODE_FILE_FLAGS_H
#define UTIL_LINUX_LSFD_DECODE_FILE_FLAGS_H

#include "buffer.h"

void lsfd_decode_file_flags(struct ul_buffer *buf, int flags);

#endif /* UTIL_LINUX_LSFD_DECODE_FILE_FLAGS_H */
