/*
 * Copyright (C) 2011 Karel Zak <kzak@redhat.com>
 *
 * This file may be redistributed under the terms of the
 * GNU Lesser General Public License.
 */
#ifndef UTIL_LINUX_MATCH_H
#define UTIL_LINUX_MATCH_H

extern int match_fstype(const char *type, const char *pattern);

#endif /* UTIL_LINUX_MATCH_H */
